/*********************************************************************
*                SEGGER Microcontroller GmbH & Co. KG                *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2017 SEGGER Microcontroller GmbH & Co. KG         *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: 4.38                                             *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : ventry.s
Purpose : Vector table for the E31/E51 RISC-V core
--------  END-OF-HEADER  ---------------------------------------------
*/

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

#ifndef   USE_ROM_VECTOR_TABLE          // Modification to this value also requires modification to RTOSInit.c! Consider using project options instead.
  #define USE_ROM_VECTOR_TABLE     (1)  // Configures ISR handling for vectored mode. A ROM vector table is used.
#endif

/*********************************************************************
*
*     Const defines
*
**********************************************************************
*/

#if (__riscv_xlen == 64)
  #define STORE     sd
  #define LOAD      ld
  #define REGBYTES   8
#else
  #define STORE     sw
  #define LOAD      lw
  #define REGBYTES   4
#endif

#define NUM_CALLER_SAVE_REGISTERS (16)  // ra, t0-t2, a0-a7, t3-t6

/*********************************************************************
*
*     Publics
*
**********************************************************************
*/

#if (USE_ROM_VECTOR_TABLE != 0)
        .global vtrap_entry
        .type   vtrap_entry, function  
#else
        .global trap_entry
        .type   trap_entry, function  
#endif

/*********************************************************************
*
*       MACROS
*
**********************************************************************
*/

      .macro TRAP_ENTRY
        addi  sp, sp, -((16 * REGBYTES) + 16)
        STORE ra,  0 * REGBYTES(sp)
        STORE t0,  1 * REGBYTES(sp)
        STORE t1,  2 * REGBYTES(sp)
        STORE t2,  3 * REGBYTES(sp)
        STORE a0,  4 * REGBYTES(sp)
        STORE a1,  5 * REGBYTES(sp)
        STORE a2,  6 * REGBYTES(sp)
        STORE a3,  7 * REGBYTES(sp)
        STORE a4,  8 * REGBYTES(sp)
        STORE a5,  9 * REGBYTES(sp)
        STORE a6, 10 * REGBYTES(sp)
        STORE a7, 11 * REGBYTES(sp)
        STORE t3, 12 * REGBYTES(sp)
        STORE t4, 13 * REGBYTES(sp)
        STORE t5, 14 * REGBYTES(sp)
        STORE t6, 15 * REGBYTES(sp)
        csrr  a0, mepc
        STORE a0, 16 * REGBYTES(sp)
      .endm

      .macro TRAP_EXIT              
        LOAD  a0, 16 * REGBYTES(sp)
        csrw  mepc, a0

        // Remain in M-mode after mret
        li    t0, 0x1800
        csrs  mstatus, t0

        LOAD  ra,  0 * REGBYTES(sp)
        LOAD  t0,  1 * REGBYTES(sp)
        LOAD  t1,  2 * REGBYTES(sp)
        LOAD  t2,  3 * REGBYTES(sp)
        LOAD  a0,  4 * REGBYTES(sp)
        LOAD  a1,  5 * REGBYTES(sp)
        LOAD  a2,  6 * REGBYTES(sp)
        LOAD  a3,  7 * REGBYTES(sp)
        LOAD  a4,  8 * REGBYTES(sp)
        LOAD  a5,  9 * REGBYTES(sp)
        LOAD  a6, 10 * REGBYTES(sp)
        LOAD  a7, 11 * REGBYTES(sp)
        LOAD  t3, 12 * REGBYTES(sp)
        LOAD  t4, 13 * REGBYTES(sp)
        LOAD  t5, 14 * REGBYTES(sp)
        LOAD  t6, 15 * REGBYTES(sp)
              
        addi  sp, sp, (16 * REGBYTES) + 16
        mret
      .endm

/*********************************************************************
*
*     CODE segment
*
**********************************************************************
*/

#if (USE_ROM_VECTOR_TABLE != 0)
/********** ROM Vector table ****************************************/
  .section .text.trap_entry, "ax", %progbits
  .align 8
vtrap_entry:  
  j SyncTrap_Handler
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
  .align 2
  j MSI_Handler
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
  .align 2
  j MTI_Handler
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
  .align 2
  j MEI_Handler
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
  .align 2
  j .  // reserved
   .align 2
  j .  // reserved
  .align 2
  j LI0_Handler
  .align 2
  j LI1_Handler
  .align 2
  j LI2_Handler
  .align 2
  j LI3_Handler
  .align 2
  j LI4_Handler
  .align 2
  j LI5_Handler
  .align 2
  j LI6_Handler
  .align 2
  j LI7_Handler
  .align 2
  j LI8_Handler
  .align 2
  j LI9_Handler
  .align 2
  j LI10_Handler
  .align 2
  j LI11_Handler
  .align 2
  j LI12_Handler
  .align 2
  j LI13_Handler
  .align 2
  j LI14_Handler
  .align 2
  j LI15_Handler
 
/*** Low-level IRQ handlers *********************************************************************/  
SyncTrap_Handler:
  TRAP_ENTRY
  csrr a0, mcause
  csrr a1, mepc
  jal  OS_ISR_SynchronousTrap
  j    .  // Execution shall halt in OS_ISR_SynchronousTrap(). If execution returns from that BSP function, we halt here.
 
MSI_Handler:
  TRAP_ENTRY
  jal  OS_ISR_Software
  TRAP_EXIT

MTI_Handler:
  TRAP_ENTRY
  jal OS_ISR_Timer
  TRAP_EXIT

MEI_Handler:
  TRAP_ENTRY
  jal OS_ISR_External
  TRAP_EXIT

LI0_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local0
  TRAP_EXIT

LI1_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local1
  TRAP_EXIT

LI2_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local2
  TRAP_EXIT
  
LI3_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local3
  TRAP_EXIT

LI4_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local4
  TRAP_EXIT

LI5_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local5
  TRAP_EXIT

LI6_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local6
  TRAP_EXIT

LI7_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local7
  TRAP_EXIT

LI8_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local8
  TRAP_EXIT

LI9_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local9
  TRAP_EXIT

LI10_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local10
  TRAP_EXIT

LI11_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local11
  TRAP_EXIT

LI12_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local12
  TRAP_EXIT

LI13_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local13
  TRAP_EXIT

LI14_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local14
  TRAP_EXIT

LI15_Handler:
  TRAP_ENTRY
  jal OS_ISR_Local15
  TRAP_EXIT

/*** Weak high-level IRQ handler, defaulting to OS_ISR_Undefined(). Can be overwritten in "C" ***/  
.weak OS_ISR_SynchronousTrap  // Implemented in RTOSInit.c
OS_ISR_SynchronousTrap:
.weak OS_ISR_Software         // Implemented in RTOSInit.c
OS_ISR_Software:
.weak OS_ISR_Timer            // Implemented in RTOSInit.c
OS_ISR_Timer:
.weak OS_ISR_External         // Implemented in RTOSInit.c, except for simulator project
OS_ISR_External:
.weak OS_ISR_Local0
OS_ISR_Local0:
.weak OS_ISR_Local1
OS_ISR_Local1:
.weak OS_ISR_Local2
OS_ISR_Local2:
.weak OS_ISR_Local3
OS_ISR_Local3:
.weak OS_ISR_Local4           // Implemented in RTOSInit.c, except for simulator project
OS_ISR_Local4:
.weak OS_ISR_Local5
OS_ISR_Local5:
.weak OS_ISR_Local6
OS_ISR_Local6:
.weak OS_ISR_Local7
OS_ISR_Local7:
.weak OS_ISR_Local8
OS_ISR_Local8:
.weak OS_ISR_Local9
OS_ISR_Local9:
.weak OS_ISR_Local10
OS_ISR_Local10:
.weak OS_ISR_Local11
OS_ISR_Local11:
.weak OS_ISR_Local12
OS_ISR_Local12:
.weak OS_ISR_Local13
OS_ISR_Local13:
.weak OS_ISR_Local14
OS_ISR_Local14:
.weak OS_ISR_Local15
OS_ISR_Local15:
  TRAP_ENTRY
  jal OS_ISR_Undefined        // Implemented in RTOSInit.c
  TRAP_EXIT


#else  // (USE_ROM_VECTOR_TABLE == 0)


       .section .text.trap_entry, "ax", %progbits
       .align 2
trap_entry:  
        addi  sp, sp, -16 * REGBYTES
        STORE ra,  0 * REGBYTES(sp)
        STORE t0,  1 * REGBYTES(sp)
        STORE t1,  2 * REGBYTES(sp)
        STORE t2,  3 * REGBYTES(sp)
        STORE a0,  4 * REGBYTES(sp)
        STORE a1,  5 * REGBYTES(sp)
        STORE a2,  6 * REGBYTES(sp)
        STORE a3,  7 * REGBYTES(sp)
        STORE a4,  8 * REGBYTES(sp)
        STORE a5,  9 * REGBYTES(sp)
        STORE a6, 10 * REGBYTES(sp)
        STORE a7, 11 * REGBYTES(sp)
        STORE t3, 12 * REGBYTES(sp)
        STORE t4, 13 * REGBYTES(sp)
        STORE t5, 14 * REGBYTES(sp)
        STORE t6, 15 * REGBYTES(sp)
        
        csrr  a0, mcause
        csrr  a1, mepc
        call  OS_IRQ_Handler // Implemented in RTOSInit.c
        csrw  mepc, a0
        
        // Remain in M-mode after mret
        li    t0, 0x1800
        csrs  mstatus, t0

        LOAD  ra,  0 * REGBYTES(sp)
        LOAD  t0,  1 * REGBYTES(sp)
        LOAD  t1,  2 * REGBYTES(sp)
        LOAD  t2,  3 * REGBYTES(sp)
        LOAD  a0,  4 * REGBYTES(sp)
        LOAD  a1,  5 * REGBYTES(sp)
        LOAD  a2,  6 * REGBYTES(sp)
        LOAD  a3,  7 * REGBYTES(sp)
        LOAD  a4,  8 * REGBYTES(sp)
        LOAD  a5,  9 * REGBYTES(sp)
        LOAD  a6, 10 * REGBYTES(sp)
        LOAD  a7, 11 * REGBYTES(sp)
        LOAD  t3, 12 * REGBYTES(sp)
        LOAD  t4, 13 * REGBYTES(sp)
        LOAD  t5, 14 * REGBYTES(sp)
        LOAD  t6, 15 * REGBYTES(sp)
        
        addi  sp, sp, 16 * REGBYTES
        mret

#endif  // USE_ROM_VECTOR_TABLE

/********************************************************************/

        .end

/****** End Of File *************************************************/
